/*
 *  DomRC5 Arduino Library, adapted from RC5 library
 *  Guy Carpenter, Clearwater Software - 2013
 *  Dominic Beesley - 2015 make interrupt driven
 *
 *  Licensed under the BSD2 license, see LICENSE for details.
 *
 *  All text above must be included in any redistribution.
 *  
 *  like plain Rc5 library but uses pin interrupts instead of polling to read the 
 *  IR pin - this means that calls to read need not be in a tight loop
 */

#ifndef RC5_h
#define RC5_h

#include <Arduino.h>

class DomRC5
{

private:
    void doRead(void);
    unsigned char pin;
    unsigned char state;
    unsigned long time0;
    int lastValue;
    unsigned int bits;
    unsigned int command;

    volatile unsigned int message;
    volatile unsigned int ready;
    void reset();
    void decodeEvent(unsigned char event);
    void decodePulse(unsigned char signal, unsigned long period);
    friend void DomRC5_ISR(void);
 public:

    DomRC5(unsigned char pin);
    bool read(unsigned int *message);
    bool read(unsigned char *toggle, unsigned char *address, unsigned char *command);
};

#endif // RC5_h

