

#define ARM_MATH_CM4
#include "arm_math.h"
#include "utils.h"
#include "DomRGBScreen.h"
#include "DomRC5.h"

#define PDB_CH0C1_TOS 0x0100
#define PDB_CH0C1_EN 0x01


#define pinTest 16
#define pinFrame 18
#define pinColourAdj 19


#define FRAME_STEP 2
#define FRAME_H 32
#define FRAME_W (96)
#define FRAME_W_SAMP (FRAME_W * FRAME_STEP)
#define SAMPLE_FREQ (FRAME_W_SAMP * 400)
#define CLOCKS_PER_SAMPLE ((F_BUS / 1 / 1 / (SAMPLE_FREQ))  -1)

#define THETA_PER_SAMPLE (50)
#define THETA_PER_SAMPLE_16 (THETA_PER_SAMPLE<<16)
//((256.0 * 65536.0 * 15000.0) / SAMPLE_FREQ)

#define LPLL_MU (FRAME_W_SAMP / (2 * PI))
#define LPLL_BETA ((float)VARIANT_MCK / (float)(FRAME_W_SAMP * CLOCKS_PER_SAMPLE * (1 + CLOCKS_PER_SAMPLE)))
#define LPLL_T 100
#define LPLL_x 0.05
#define LPLL_GAIN 10
#define LPLL_OFFS 3

#define CPLL_MU (1024)
#define CPLL_BETA ((float)15000 / (float)80)
#define CPLL_T 0.1
#define CPLL_x 0.4
#define CPLL_GAIN 3

//map A pins to analogue channels
static const uint8_t channel2sc1a[] = {
  5, 14, 8, 9, 13, 12, 6, 7, 15, 4,
  0, 19, 3, 21, 26, 22
};

#define LUMA_DELAY 44


//BUTTR14B + delay stages
const int IIR_BS_LUMA_STAGES = 7;
arm_biquad_casd_df1_inst_q15 iir_bs_luma_S;
q15_t iir_bs_luma_coeffs[IIR_BS_LUMA_STAGES * 6] = {
0x1269,0x0000,0xf36e,0x1269,0x44eb,0xc660,
0x3fff,0x0000,0xd44c,0x3fff,0x0847,0xc7a4,
0x3fff,0x0000,0xd44c,0x3fff,0x3ad5,0xd100,
0x3fff,0x0000,0xd44c,0x3fff,0x0d19,0xd376,
0x3fff,0x0000,0xd44c,0x3fff,0x2ef2,0xd884,
0x3fff,0x0000,0xd44c,0x3fff,0x1637,0xda5d,
0x3fff,0x0000,0xd44c,0x3fff,0x2226,0xdbfc
};
q15_t iir_bs_luma_state[IIR_BS_LUMA_STAGES * 4];
int iir_bs_luma_shift = 1;

/*
 // no filter
const int IIR_BS_LUMA_STAGES = 1;
arm_biquad_casd_df1_inst_q15 iir_bs_luma_S;
q15_t iir_bs_luma_coeffs[IIR_BS_LUMA_STAGES * 6] = {
0,0,0x4000,0,0,0
};
q15_t iir_bs_luma_state[IIR_BS_LUMA_STAGES * 4];
int iir_bs_luma_shift = 1;
*/


//BUTTR08B
const int IIR_LP_UV_STAGES = 4;
arm_biquad_casd_df1_inst_q15 iir_lp_U_S;
arm_biquad_casd_df1_inst_q15 iir_lp_V_S;
q15_t iir_lp_uv_coeffs[IIR_LP_UV_STAGES * 6] = {
0x004a,0x0000,0x0094,0x004a,0x77b2,0xc548,
0x00bd,0x0000,0x017b,0x00bd,0x6f37,0xcdfa,
0x0101,0x0000,0x0203,0x0101,0x697f,0xd3d7,
0x010f,0x0000,0x021f,0x010f,0x66a4,0xd6c5
};
q15_t iir_lp_U_state[IIR_LP_UV_STAGES * 4];
q15_t iir_lp_V_state[IIR_LP_UV_STAGES * 4];
int iir_lp_uv_shift = 1;


//BUTTR16A
const int IIR_HP_CHROMA_STAGES = 8;
arm_biquad_casd_df1_inst_q15 iir_hp_chroma_S;
q15_t iir_hp_chroma_coeffs[IIR_HP_CHROMA_STAGES * 6] = {
0x0288,0x0000,0x0000,0xfd78,0x45a9,0xc512,
0x207d,0x0000,0x0000,0xdf83,0x0f7e,0xc60e,
0x070b,0x0000,0x0000,0xf8f5,0x3dd2,0xcdce,
0x1bea,0x0000,0x0000,0xe416,0x1240,0xcfe6,
0x0b10,0x0000,0x0000,0xf4f0,0x34a8,0xd471,
0x1725,0x0000,0x0000,0xe8db,0x185b,0xd665,
0x0eb4,0x0000,0x0000,0xf14c,0x2ab5,0xd87b,
0x11f6,0x0000,0x0000,0xee0a,0x20df,0xd945
};
q15_t iir_hp_chroma_state[IIR_HP_CHROMA_STAGES * 4];
int iir_hp_chroma_shift = 1;

volatile unsigned char buf_raw_top_ready = 0;
volatile unsigned char buf_raw_bottom_ready = 0;
q15_t buf_raw[FRAME_W_SAMP * 2];
q15_t buf[FRAME_W_SAMP];
q15_t delay_buffer[LUMA_DELAY];
q15_t buf2[FRAME_W_SAMP];
q15_t buf_chroma[FRAME_W_SAMP];
q15_t buf_I_raw[FRAME_W_SAMP];
q15_t buf_Q_raw[FRAME_W_SAMP];
q15_t buf_I[FRAME_W_SAMP];
q15_t buf_Q[FRAME_W_SAMP];


#define LIM(x) ((x>255)?255:(x<0)?0:x)

#define IR_PIN 22
DomRC5 rc5(IR_PIN);

unsigned char lim(int r) {
  if (r < 0)    
    return 0;  
  else if (r > 255)
    return 255;
  else
    return r;
}

signed char lim2(int r) {
  if (r < -127)    
    return -127;  
  else if (r > 127)
    return 127;
  else
    return r;
}


unsigned char gam[256];
signed char sin_tab[256];

int I_theta = 0;
int Q_theta = 0;


//Used in the timing routine - 65536 * the number of ticks for the 
//video sample rate
volatile int pdb_ticks_16_next = 0;
volatile int pdb_ticks_16_live = 0;
volatile int pdb_ticks_16_last = 0;
//when this overflows to 65536 add an extra tick
volatile int pdb_fraction_acc = 0;



/*
  ADC_CFG1_ADIV(2)         Divide ratio = 4 (F_BUS = 48 MHz => ADCK = 12 MHz)
  ADC_CFG1_MODE(2)         Single ended 16 bit mode
  ADC_CFG1_ADLSMP          Shng sample time
*/
#define ADC_CONFIG1 (ADC_CFG1_ADIV(0) | ADC_CFG1_MODE(1) | ADC_CFG1_ADICLK(0))

/*
  ADC_CFG2_MUXSEL          Select channels ADxxb
  ADC_CFG2_ADLSTS(3)       Shortest long sample time
*/
#define ADC_CONFIG2 (ADC_CFG2_MUXSEL | ADC_CFG2_ADLSTS(3) | ADC_CFG2_ADHSC)

void adcInit() {
  ADC0_CFG1 = ADC_CONFIG1;
  ADC0_CFG2 = ADC_CONFIG2;
  // Voltage ref vcc, hardware trigger, DMA
  ADC0_SC2 = ADC_SC2_REFSEL(0) | ADC_SC2_ADTRG | ADC_SC2_DMAEN;


  adcCalibrate();
  Serial.println("calibrated");

  // Enable averaging, 32 samples
  ADC0_SC3 = ADC_SC3_AVGE | ADC_SC3_AVGS(3);


  // Enable ADC interrupt, configure pin
  ADC0_SC1A = 23; //channel2sc1a[3]; //ADC_SC1_AIEN | 
  //NVIC_ENABLE_IRQ(IRQ_ADC0);
}

void adcCalibrate() {
  uint16_t sum;

  // Begin calibration
  ADC0_SC3 = ADC_SC3_CAL;
  // Wait for calibration
  while (ADC0_SC3 & ADC_SC3_CAL);

  // Plus side gain
  sum = ADC0_CLPS + ADC0_CLP4 + ADC0_CLP3 + ADC0_CLP2 + ADC0_CLP1 + ADC0_CLP0;
  sum = (sum / 2) | 0x8000;
  ADC0_PG = sum;

  // Minus side gain (not used in single-ended mode)
  sum = ADC0_CLMS + ADC0_CLM4 + ADC0_CLM3 + ADC0_CLM2 + ADC0_CLM1 + ADC0_CLM0;
  sum = (sum / 2) | 0x8000;
  ADC0_MG = sum;
}

/*
  PDB_SC_TRGSEL(15)        Select software trigger
  PDB_SC_PDBEN             PDB enable
  PDB_SC_PDBIE             Interrupt enable
  PDB_SC_CONT              Continuous mode
  PDB_SC_PRESCALER(7)      Prescaler = 1
  PDB_SC_MULT(1)           Prescaler multiplication factor = 1
  PDB_LDMOD(1)             load at next MOD
*/
#define PDB_CONFIG (PDB_SC_TRGSEL(15) | PDB_SC_PDBEN | PDB_SC_PDBIE \
  | PDB_SC_CONT | PDB_SC_PRESCALER(0) | PDB_SC_MULT(0) | PDB_SC_LDMOD(1))

// 48 MHz / 1 / 1 / 1 Hz = 37500
#define PDB_PERIOD CLOCKS_PER_SAMPLE

void pdbInit() {
  // Enable PDB clock
  SIM_SCGC6 |= SIM_SCGC6_PDB;
  // Timer period
  PDB0_MOD = PDB_PERIOD;
  pdb_ticks_16_live = PDB_PERIOD << 16;
  pdb_ticks_16_next = PDB_PERIOD << 16;
  // Interrupt delay
  PDB0_IDLY = 0;
  // Enable pre-trigger
  PDB0_CH0C1 = PDB_CH0C1_TOS | PDB_CH0C1_EN;
  // PDB0_CH0DLY0 = 0;
  PDB0_SC = PDB_CONFIG | PDB_SC_LDOK;
  // Software trigger (reset and restart counter)
  PDB0_SC |= PDB_SC_SWTRIG;
  // Enable interrupt request
  NVIC_ENABLE_IRQ(IRQ_PDB);
}

void dmaInit() {
  // Enable DMA, DMAMUX clocks
  SIM_SCGC7 |= SIM_SCGC7_DMA;
  SIM_SCGC6 |= SIM_SCGC6_DMAMUX;

  // Use default configuration
  DMA_CR = 0;

  // Source address
  DMA_TCD1_SADDR = &ADC0_RA;
  // Don't change source address
  DMA_TCD1_SOFF = 0;
  DMA_TCD1_SLAST = 0;
  // Destination address
  DMA_TCD1_DADDR = buf_raw;
  // Destination offset (2 byte)
  DMA_TCD1_DOFF = 2;
  // Restore destination address after major loop
  DMA_TCD1_DLASTSGA = -FRAME_W_SAMP * 4;
  // Source and destination size 16 bit
  DMA_TCD1_ATTR = DMA_TCD_ATTR_SSIZE(1) | DMA_TCD_ATTR_DSIZE(1);
  // Number of bytes to transfer (in each service request)
  DMA_TCD1_NBYTES_MLNO = 2;
  // Set loop counts
  DMA_TCD1_CITER_ELINKNO = FRAME_W_SAMP * 2;
  DMA_TCD1_BITER_ELINKNO = FRAME_W_SAMP * 2;
  // Enable interrupt (end-of-major loop)
  DMA_TCD1_CSR = DMA_TCD_CSR_INTMAJOR | DMA_TCD_CSR_INTHALF;

  // Set ADC as source (CH 1), enable DMA MUX
  DMAMUX0_CHCFG1 = DMAMUX_DISABLE;
  DMAMUX0_CHCFG1 = DMAMUX_SOURCE_ADC0 | DMAMUX_ENABLE;

  // Enable request input signal for channel 1
  DMA_SERQ = 1;

  // Enable interrupt request
  NVIC_ENABLE_IRQ(IRQ_DMA_CH1);
}


volatile int pixelPtr = 0;
int vid_val;



int bank = 0;

float c_pll_C = 0.0;
int c_err_V = 0;

const int r_v = (int)(128.0 * 1.13983);
const int g_u = (int)(128.0 * 0.39465);
const int g_v = (int)(128.0 * 0.58060);
const int b_u = (int)(128.0 * 2.03211);

int prev_Y = 0, Y = 0;    

int y = 0;

int test_mode = 0;
int last_test_button = 0;

byte sync = 0;
short samples_since_sync = 0;
byte frame_sync = 0;
byte lines_synce_frame_sync = 0;

int black = 0;

int synctip = 0;
//error as detected by last line sync
//C voltage in RRC circuit
float l_pll_C = 0.0;
//pll ticks - control error voltage in ticks * 256 after filtering
int l_err_V = 0;

int tot_sync_err = 0;
int tot_sync_size = 0;
int tot_sync_samples = 0;
int sync_count = 0;
int syncs_per_frame = 0;
#define MAX_FRAME_SYNCS 50
int frame_line_syncs[MAX_FRAME_SYNCS];
#define MIN_SYNC_SIZE (2 * FRAME_STEP)

#define SHOW_SYNC(x, Y, I)       if (test_mode & 2)     \
      {                                                 \
        buf[x] = Y;                                   \   
        if (x >= 1) buf[x - 1] = Y;                     \
      }                                                 \

//user controls
int bri = 100;
int cerr_corr_sign = 1;
int sat = 128;
int hue = 0;
int sync_lock = 1;
int black_lock = 1;

void doLineSyncs() {
  byte had_frame_sync = 0;
  int sync_slice = ((synctip >> 5) - 280);
  //get syncs
  for (int x=0; x < FRAME_W_SAMP; x++) {
    int v = buf[x] >> 1;    

    int sync_size = v - sync_slice;
    if (sync_size > 0)
    {
      SHOW_SYNC(x, 1800, +500);

      sync = 1;
      int offs = ((x > (FRAME_W_SAMP >> 1))?x-FRAME_W_SAMP:x);
      tot_sync_err += offs * sync_size;
      tot_sync_size += sync_size;
      tot_sync_samples ++;
    } else {

      // end of sync, check is big enough sync and do line sync adjustments
      if (sync == 1 && tot_sync_samples > MIN_SYNC_SIZE)
      {
        SHOW_SYNC(x, 1000, +500);
        sync_count++;
        samples_since_sync = 0;
      }
      else
        samples_since_sync ++;
      
      sync = 0;
      tot_sync_samples = 0;

      if (lines_synce_frame_sync > 30 && samples_since_sync > (FRAME_W_SAMP + (FRAME_W_SAMP >> 1))) {
        SHOW_SYNC(x, 1000, -800);
        had_frame_sync = 1;
        if (frame_sync == 0 && lines_synce_frame_sync == 31) {

          Serial.println(lines_synce_frame_sync);
          frame_sync = 1;
          //TODO: Make a proper flywheel frame sync
        } 
      } else {
        frame_sync = 0;
      }
    }
  
    if (v > synctip >> 5) {
      SHOW_SYNC(x, 1000, -500);
      synctip = (synctip >> 1) + (v << 4);
    }
      
    synctip -=6;
  }

  if (sync_lock && sync_count > 0) {

    float avg_sync_err = LPLL_GAIN * (((float)tot_sync_err / (float)tot_sync_size) - LPLL_OFFS);

    if (syncs_per_frame < MAX_FRAME_SYNCS)
      frame_line_syncs[syncs_per_frame] = avg_sync_err;
    syncs_per_frame++;

    
    l_pll_C += ((float)avg_sync_err - l_pll_C) / ((float)LPLL_T * 400);
    l_err_V = (int)65536.0 * (l_pll_C + (float)LPLL_x * ((float)avg_sync_err - l_pll_C));

    if (l_pll_C > 15) l_pll_C = 15;
    if (l_pll_C < -15) l_pll_C = -15;

    if (l_err_V > (15 << 16)) l_err_V = 15 << 16;
    if (l_err_V < (-15 << 16)) l_err_V = -15 << 16;
    pdb_ticks_16_next = (PDB_PERIOD << 16) + l_err_V;
    sync_count = 0;
    tot_sync_err = 0;
    tot_sync_size = 0;
    tot_sync_samples = 0;
  }

  if (frame_sync || had_frame_sync)
    lines_synce_frame_sync = 0;
  else
    lines_synce_frame_sync ++;
    
}



void setup() {
  Serial.begin(115200);

  delay(250);

  for (int i = 0; i < 256; i++) {
    gam[i] = lim((int)((float)(1<<MAX_PLANES) * pow((float)(i)/256.0, 2.1)));    
    //gam[i] = (float)(1<<MAX_PLANES) * (float)(i)/256.0;
    sin_tab[i] = lim2((int)(127.0 * sin(2.0*PI*(float)i/256.0)));
  }


  _domRGBScreen.begin();


  pinMode(pinTest, OUTPUT);
  pinMode(pinFrame, OUTPUT);
  pinMode(pinColourAdj, INPUT);


  adcInit();
  pdbInit();
  dmaInit();
  
  for (int i = 0; i < 2; i++) {
    _domRGBScreen.flip(i);
    unsigned char *scr = _domRGBScreen.getScreenMemBase();
    for (int j = 0; j < SCR_WIDTH * SCR_HEIGHT; j ++) {
      if (i) {
        scr[0] = gam[((j&0x1F) << 3)];
        scr[1] = 0;
        scr[2] = 0;
      } else {
        int c = j / 64;
        if (j & 4) {
          scr[0] = 0;
          scr[1] = 0;
          scr[2] = 1 << (c & 7);        
        } else {
          scr[0] = 0;
          scr[1] = gam[(c & 3) << 6];                  
          scr[2] = 0;
        }
      }
      scr+=4;
    }
  }

  int i = 0;
  while (i < 3) {
    for (int j = 0; j < 50; j++) {
      _domRGBScreen.waitVSync();
    }
    _domRGBScreen.flip(i++);
  }

  _domRGBScreen.flip(0);
  memset(_domRGBScreen.getScreenMemBase(), 0, SCR_WIDTH * SCR_HEIGHT * 4);
  _domRGBScreen.flip(1);
  memset(_domRGBScreen.getScreenMemBase(), 0, SCR_WIDTH * SCR_HEIGHT * 4);

  arm_biquad_cascade_df1_init_q15(&iir_bs_luma_S, IIR_BS_LUMA_STAGES, iir_bs_luma_coeffs, iir_bs_luma_state, iir_bs_luma_shift);

  arm_biquad_cascade_df1_init_q15(&iir_lp_U_S, IIR_LP_UV_STAGES, iir_lp_uv_coeffs, iir_lp_U_state, iir_lp_uv_shift);
  arm_biquad_cascade_df1_init_q15(&iir_lp_V_S, IIR_LP_UV_STAGES, iir_lp_uv_coeffs, iir_lp_V_state, iir_lp_uv_shift);
  

  arm_biquad_cascade_df1_init_q15(&iir_hp_chroma_S, IIR_HP_CHROMA_STAGES, iir_hp_chroma_coeffs, iir_hp_chroma_state, iir_hp_chroma_shift);

  I_theta = 64 << 16;
  Q_theta = 0 << 16;

  memset(iir_bs_luma_state, 0, sizeof(q15_t) * 4 * IIR_BS_LUMA_STAGES);

  lines_synce_frame_sync = 0;
  test_mode = -1;

  bri = 100;
}



int show_ui = 0;  //when >0 show UI
unsigned char last_toggle;

void showUI(unsigned char *scr) 
{
  
}

void checkIR() {
  unsigned char toggle;
  unsigned char address;
  unsigned char command;

  int test_button = digitalRead(pinColourAdj);
  if (last_test_button == 0 && test_button == 1)
  {
    test_mode ++;
  }
  last_test_button = test_button;

  
  //infrared
  if (rc5.read(&toggle, &address, &command))
  {
    Serial.print(address);
    Serial.print(",");
    Serial.print(command);
    Serial.println(toggle ? " (t) " : "");

    if (address == 0) {
      switch(command) {
        case 16:
          bri+=10;
          if (bri > 200) bri = 200;
          break;
        case 17:
          bri-=10;
          if (bri < 0) bri = 0;
          break;
        case 32:
          sat ++;
          if (sat > 255) sat = 255;
          break;
        case 33:
          sat --;
          if (sat < 0) sat = 0;
          break;
        case 22:
          if (toggle != last_toggle)
            test_mode++;
          break;
        case 14:
          if (toggle != last_toggle) {
            if (cerr_corr_sign == -1)
              cerr_corr_sign = 0;
            else if (cerr_corr_sign == 0)
              cerr_corr_sign = 1;
            else
              cerr_corr_sign = -1;
          }
          break;
        case 1:
          hue+=10;
          if (hue > 200) hue = 200;
          break;
        case 2:
          hue-=10;
          if (hue < -200) hue = -200;
          break;
        case 55:
          if (toggle != last_toggle)
          {
            if (sync_lock)
              sync_lock = 0;
            else
              sync_lock = 1;
          }
          break;
        case 54:
          if (toggle != last_toggle)
          {
            if (black_lock)
              black_lock = 0;
            else
              black_lock = 1;
          }
          break;
      }
    }
    last_toggle = toggle;
  }
  else
    Serial.println();

}

int c_line_err = 0;

void doChromaPLL() {
       Serial.print("Q=");
      int st = micros();
      int Q_theta_tmp = Q_theta + ((THETA_PER_SAMPLE_16) + c_line_err + c_err_V) * FRAME_W_SAMP / 4;
      int Qd = 0;
      for (int i = FRAME_W_SAMP / 2 ; i < FRAME_W_SAMP - 10; i ++ ) {
        int QdInst = -(((int)(buf_chroma[i] >> 2) * sin_tab[(Q_theta_tmp >> 16) & 255]) >> 2) * CPLL_GAIN;
        //int QdInst = 0;
        Qd = QdInst + ((Qd * 255) >> 8);
  
#define QLIM 256000
        if (Qd > QLIM)
          Qd = QLIM;
        else if (Qd < -QLIM)
          Qd = -QLIM;
  
        int Qd2 = Qd / 256;
          //Serial.print(" ");
          //Serial.print(Qd2);
        
        if (i > FRAME_W_SAMP / 2) {
          c_pll_C += ((float)Qd2 - c_pll_C) / (CPLL_T * SAMPLE_FREQ);
  
          c_err_V =  c_pll_C + CPLL_x * ((float)Qd2 - c_pll_C); 
        }
        Q_theta_tmp += (THETA_PER_SAMPLE_16) + c_line_err + c_err_V;
      }
      Serial.println();
      Serial.println(micros() - st);
}

//float CTOLFACTOR = (float)THETA_PER_SAMPLE / (float)PDB_PERIOD;
//float pdb_avg = (float)(PDB_PERIOD << 16);

void loop() { 

  
  while (!buf_raw_top_ready && !buf_raw_bottom_ready) ;

//  int pdb_ticks_16 = pdb_ticks_16_last;
//  pdb_avg = 0.999 * pdb_avg + 0.001 * pdb_ticks_16;
//  Serial.print("line ticks=");
//  Serial.print(pdb_ticks_16);
//  int c_line_err = cerr_corr_sign * (int) (((float)pdb_ticks_16 - pdb_avg) * CTOLFACTOR);
//  Serial.print(" ");
//  Serial.print(c_line_err);
//  Serial.print(" ");
//  Serial.print(c_err_V);
//  Serial.println();

  digitalWrite(pinFrame, 1);

  unsigned int off = 0;
  if (buf_raw_top_ready)
    off = FRAME_W_SAMP;

  buf_raw_top_ready = buf_raw_bottom_ready = 0;

  arm_biquad_cascade_df1_fast_q15(&iir_bs_luma_S, buf_raw + off, buf2, FRAME_W_SAMP);
  arm_biquad_cascade_df1_fast_q15(&iir_hp_chroma_S, buf_raw + off, buf_chroma, FRAME_W_SAMP);

  memcpy(buf, delay_buffer , LUMA_DELAY * 2);
  memcpy(buf + LUMA_DELAY, buf2, (FRAME_W_SAMP - LUMA_DELAY) *2);
  memcpy(delay_buffer, buf2 + (FRAME_W_SAMP - LUMA_DELAY), LUMA_DELAY * 2); 

  unsigned char * screen_base = _domRGBScreen.getScreenMemBase();


  //black level
  if (black_lock && y != 0) {
    for (int x=2; x < 5; x++) {
        black = ((black * 31) >> 5) + (((int)buf[x]));      
        //black = (((int)buf[x])) << 5;
    }
  }


  doLineSyncs();

  
/*
  if (y == 20) {
    for (int i = FRAME_W_SAMP - 40; i < FRAME_W_SAMP + 40; i++)
    {
      Serial.print(",");
      Serial.print(buf[i % FRAME_W_SAMP]);  
    }
    Serial.println();
  }
*/
  // apply black level
  for (int x=0; x < FRAME_W_SAMP; x++) {
    buf[x] =  (((black) - (buf[x] <<5)) >>8) - bri;
    //buf[x] =  ((black - ((int)buf[x]) << 5) >> 7) - 100;
  }



  uint8_t calc_speeds = 0;

  y++;
  if (y >=FRAME_H || frame_sync) {
    y = 0;

    showUI(screen_base);
  
    _domRGBScreen.waitLSync();
    _domRGBScreen.flip(bank++);

    //Serial.println((float)c_err_V / (float)65536);
  
    calc_speeds = 1;
  }



  if (y == 0) {
    int Qd = 0;
    //decode IQ and update pll
    for (int x=0; x < FRAME_W_SAMP; x++) {
      buf_I_raw[x] = (((int)(buf_chroma[x] >> 2) * sin_tab[(I_theta >> 16) & 255]) >> 2);
      int QdInst = (((int)(buf_chroma[x] >> 2) * sin_tab[(Q_theta >> 16) & 255]) >> 2);
      buf_Q_raw[x] = QdInst;
      I_theta += (THETA_PER_SAMPLE_16) + c_line_err + c_err_V;
      Q_theta += (THETA_PER_SAMPLE_16) + c_line_err + c_err_V;
              
      //int QdInst = 0;
      Qd = ((QdInst * CPLL_GAIN) + (Qd * 127)) / 128;
  
#define QLIM 3000
        if (Qd > QLIM)
          Qd = QLIM;
        else if (Qd < -QLIM)
          Qd = -QLIM;
  
        int Qd2 = Qd;
          //Serial.print(" ");
          //Serial.print(Qd2);

        if (x >= 100 && x <= 105) {
          Serial.print("Q=");
          Serial.println(Qd2);
        }
        
        if (x > FRAME_W_SAMP / 2) {
          c_pll_C += ((float)Qd2 - c_pll_C) / (CPLL_T * SAMPLE_FREQ);
  
          c_err_V =  c_pll_C + CPLL_x * ((float)Qd2 - c_pll_C); 
        }

    }

  } else {
    //decode IQ
    for (int x=0; x < FRAME_W_SAMP; x++) {
      int c = (int)(buf_chroma[x] >> 2);
      buf_I_raw[x] = ((c * sin_tab[(hue + (I_theta >> 16)) & 255]) >> 2);
      buf_Q_raw[x] = ((c * sin_tab[(hue + (Q_theta >> 16)) & 255]) >> 2);
      I_theta += (THETA_PER_SAMPLE_16) + c_line_err + c_err_V;
      Q_theta += (THETA_PER_SAMPLE_16) + c_line_err + c_err_V;
    }
  }  

  

  unsigned char *p_line = screen_base + (((y * 2) + 1) * (SCR_WIDTH * 4)) - 4;
  unsigned char *p = p_line;
  unsigned char *p2 = p_line + SCR_WIDTH * 4;;


  arm_biquad_cascade_df1_fast_q15(&iir_lp_U_S, buf_I_raw, buf_I, FRAME_W_SAMP);
  arm_biquad_cascade_df1_fast_q15(&iir_lp_V_S, buf_Q_raw, buf_Q, FRAME_W_SAMP);

  int x = 0;

  
  for (int i = 0 ; i < FRAME_W_SAMP; i+= FRAME_STEP) {

    Y = buf[i];   
    

    int I, Q;
    if (test_mode & 1) {
      I = 0;   
      Q = 0;
    } else {
      I = (sat * buf_I[i]) >> 8;
      Q = (sat * buf_Q[i]) >> 8; 
    }

    
    if (x < SCR_WIDTH) {

      if (test_mode & 4) {
        p2[0] = p[0] = 0;
        p2[1] = p[1] = 0;
      } else {
        p2[0] = p[0] = gam[LIM(Y + ((r_v * Q) >> 10))];
        p2[1] = p[1] = gam[LIM(Y - ((g_v * Q + g_u * I) >> 10))];        
      }
      p2[2] = p[2] = gam[LIM(Y + ((b_u * I) >> 10))];
      
      p-=4;
      p2-=4;
    }
    
    x++;
  }  

    
  if (calc_speeds) {
        
      //l_err_V = 0;

      float stddev = 0;
      float avg_sync = 0;
      if (!(test_mode & 2)) {
        int tot_sync = 0;
       
        for (int i = 0 ; i < syncs_per_frame; i++)
        {
          tot_sync += frame_line_syncs[i];
        }
        avg_sync = (syncs_per_frame > 0)?(float)tot_sync / (float)syncs_per_frame:0.0;
  
        float var = 0;
        for (int i = 0 ; i < syncs_per_frame; i++)
        {
          float diff = (float)frame_line_syncs[i] - avg_sync;
          var += diff * diff;
        }
  
        var = (syncs_per_frame > 0)?var / syncs_per_frame:0.0;
        stddev = sqrt(var);
      }

      Serial.print("=L====");
      Serial.print(syncs_per_frame);
      Serial.print(" ");
      Serial.print(avg_sync);
      Serial.print("+-");
      Serial.print(stddev);
      Serial.print(", C=");
      Serial.print(l_pll_C);
      Serial.print("  ");
      Serial.println((float)l_err_V / 65536.0);


      Serial.print("=C====");
      Serial.print(c_pll_C);
      Serial.print("  ");
      Serial.print(c_err_V);
      Serial.print("  ");
      Serial.println(c_line_err);

      Serial.print("bri=");
      Serial.print(bri);
      Serial.println();

/*      Serial.print("black=");
      Serial.println(black);
*/    
    syncs_per_frame = 0;

    /*Serial.print("=C=====");
    Serial.println(c_err_V);
    */

    checkIR();
   
  }

  digitalWrite(pinFrame, 0);
}


void pdb_isr() {
  // Clear interrupt flag
  PDB0_SC &= ~PDB_SC_PDBIF;

  int over = 0;
  pdb_fraction_acc += pdb_ticks_16_live & 0xFFFF;
  if (pdb_fraction_acc > 1 << 16) {
    over = 1;
    pdb_fraction_acc = pdb_fraction_acc & 0xFFFF;
  }

  PDB0_MOD = (pdb_ticks_16_live >> 16) + over;
  PDB0_SC |= PDB_SC_LDOK;

}

void dma_ch1_isr() {

  unsigned int pos = (unsigned int)DMA_TCD1_DADDR;
  
  // Clear interrupt request for channel 1
  DMA_CINT = 1;

  pdb_ticks_16_last = pdb_ticks_16_live;
  pdb_ticks_16_live = pdb_ticks_16_next;

  if (pos - (unsigned int)buf_raw < FRAME_W_SAMP) {
    buf_raw_top_ready = 1;
  } else {
    buf_raw_bottom_ready = 1;
  } 
}
